pub mod atomic;
pub mod inlining;

use std::{error::Error, fmt::Display, rc::Rc};

use crate::core::*;
use crate::func::*;

struct WrappedOrigin<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> {
    w_factory: A::Fctr,
    w_resolve: Box<dyn 'a + Fn() -> <<Ctx as Context>::T as WeakFunctor>::F<'a, Rc<A>>>,
}

trait MappableOrigin<'a, Ctx: 'a + Context>: Origin<'a, Ctx> {
    fn map<B: Mentionable<'a, Ctx>>(
        self: Rc<Self>,
        map_mentionable: impl 'a + Clone + Fn(Rc<Self::Mtbl>) -> B,
        map_factory: impl 'a + FnOnce(<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr) -> B::Fctr,
    ) -> WrappedOrigin<'a, Ctx, B>;
}

fn map_resolve<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>, B: Mentionable<'a, Ctx>>(
    resolve: impl 'a + Fn() -> <<Ctx as Context>::T as WeakFunctor>::F<'a, Rc<A>>,
    map_mentionable: impl 'a + Fn(Rc<A>) -> B,
) -> <<Ctx as Context>::T as WeakFunctor>::F<'a, Rc<B>> {
    Ctx::T::fmap(
        move |resolved| Rc::new(map_mentionable(resolved)),
        resolve(),
    )
}

impl<'a, Ctx: 'a + Context, O: ?Sized + Origin<'a, Ctx>> MappableOrigin<'a, Ctx> for O {
    fn map<B: Mentionable<'a, Ctx>>(
        self: Rc<Self>,
        map_mentionable: impl 'a + Clone + Fn(Rc<Self::Mtbl>) -> B,
        map_factory: impl 'a + FnOnce(<Self::Mtbl as Mentionable<'a, Ctx>>::Fctr) -> B::Fctr,
    ) -> WrappedOrigin<'a, Ctx, B> {
        let origin = self.clone();
        WrappedOrigin {
            w_factory: map_factory(self.factory()),
            w_resolve: Box::new(move || {
                let origin = origin.clone();
                let map_mentionable = map_mentionable.clone();
                map_resolve(
                    move || origin.resolve(),
                    map_mentionable,
                )
            }),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Origin<'a, Ctx> for WrappedOrigin<'a, Ctx, A> {
    type Mtbl = A;

    fn factory(&self) -> <Self::Mtbl as Mentionable<'a, Ctx>>::Fctr {
        self.w_factory.clone()
    }

    fn resolve(&self) -> <<Ctx as Context>::T as WeakFunctor>::F<'a, Rc<Self::Mtbl>> {
        (self.w_resolve)()
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Point<'a, Ctx, A> {
    pub fn typeless(&self) -> Point<'a, Ctx, TypelessMentionable<'a, Ctx>> {
        Point {
            point: self.point,
            origin: Rc::new(self.origin.clone().map(
                TypelessMentionable::from_mentionable,
                TypelessFactory::from_factory,
            )),
        }
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Serializable for Point<'a, Ctx, A> {
    fn serialize(&self, serializer: &mut dyn Serializer) {
        serializer.write(&self.point)
    }
}

impl<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>> Mentionable<'a, Ctx> for Point<'a, Ctx, A> {
    type Fctr = PointFactory<A::Fctr>;

    fn factory(&self) -> Self::Fctr {
        PointFactory {
            factory: self.origin.factory(),
        }
    }

    fn topology(&self) -> Hash {
        Ctx::hash(&self.point)
    }

    fn points(&self) -> Vec<Point<'a, Ctx, TypelessMentionable<'a, Ctx>>> {
        vec![self.typeless()]
    }
}

#[derive(Clone)]
pub struct PointFactory<F> {
    factory: F,
}

#[derive(Debug)]
pub enum PointParseError {
    WrongLength(usize),
}

impl Display for PointParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            PointParseError::WrongLength(length) => f.write_fmt(format_args!(
                "expected {} bytes, received {}",
                HASH_SIZE, length
            )),
        }
    }
}

impl Error for PointParseError {}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> Factory<'a, Ctx> for PointFactory<F> {
    type Mtbl = Point<'a, Ctx, F::Mtbl>;

    type ParseError = PointParseError;

    fn deserialize(
        &self,
        deserializer: &mut dyn Deserializer,
    ) -> Result<Self::Mtbl, Self::ParseError> {
        match deserializer.read_n_const::<HASH_SIZE>() {
            Ok(point) => Ok(Point {point, origin: todo!()}),
            Err(slice) => Err(PointParseError::WrongLength(slice.len())),
        }
    }
}
